#!/usr/bin/env perl
# $Header: /code/pg_trompe/replication.pl,v 1.16 2008/01/06 21:27:11 doj Exp $

use strict;
use Getopt::Long;
use DBI;
use Sys::Syslog;

use constant PRIMARY_KEY => 'nr';
use constant LO_KEY => 'lo';
use constant LO_MAGIC => '___lo___';
use constant COLUMN_MAGIC => '___column___';
use constant EMPTY_MAGIC => '___emtpy___';

my $sth;
my $sth2;
my @after_replication;

# get command line arguments

my $local_dbname;
my $local_dbuser;
my $local_dbpass;
my $local_dbhost;
my $local_dbport;
my $remote_dbname;
my $remote_dbuser;
my $remote_dbpass;
my $remote_dbhost;
my $remote_dbport;
my $syslog_fac="local6";
my $compare_db;
my $stats;

$_=GetOptions(
	      "local_dbname=s"	=> \$local_dbname,
	      "local_dbuser=s"	=> \$local_dbuser,
	      "local_dbpass=s"	=> \$local_dbpass,
	      "local_dbhost=s"	=> \$local_dbhost,
	      "local_dbport=i"	=> \$local_dbport,
	      "remote_dbname=s"	=> \$remote_dbname,
	      "remote_dbuser=s"	=> \$remote_dbuser,
	      "remote_dbpass=s"	=> \$remote_dbpass,
	      "remote_dbhost=s"	=> \$remote_dbhost,
	      "remote_dbport=i"	=> \$remote_dbport,
	      "syslog_facility=s" => \$syslog_fac,
	      "compare_db=s" => \$compare_db,
	      "stats=s" => \$stats,
	     );

openlog("pg_trompe", "pid", $syslog_fac);
$SIG{'__DIE__'} = sub {
    my $e="DIE: ".join('; ', @_);
    syslog('crit', $e);
    print STDERR "$e\n";
};
$SIG{'__WARN__'} = sub {
    my $w="WARN: ".join('; ', @_);
    syslog('warning', $w);
    print STDERR "$w\n";
};
sub mylog
{
    my $l=join(' ', @_);
    syslog('info', $l);
    print STDERR $l;
}

# connect to databases

sub connectdb
{
    my ($dbname, $dbuser, $dbpass, $dbhost, $dbport) = @_;
    my $c="dbi:Pg:dbname=$dbname".($dbhost?";host=$dbhost":'').($dbport?";port=$dbport":'');
    mylog("connect to $c $dbuser $dbpass\n");
    my $dbh=DBI->connect($c, $dbuser, $dbpass, { RaiseError => 1, AutoCommit => 0 }) or die "could not connect to $c";
    return begin_transaction($dbh);
}

sub begin_transaction
{
    my ($dbh) = @_;
    return unless ref $dbh;
    my $q="SET CONSTRAINTS ALL DEFERRED";
    $dbh->do($q) or die "could not set $q";
    return $dbh;
}

die "local_dbname not set" unless $local_dbname;
my $ldbh=connectdb($local_dbname, $local_dbuser, $local_dbpass, $local_dbhost, $local_dbport);
showstats($ldbh, "Local Database: $local_dbname") if $stats;

die "remote_dbname not set" unless $remote_dbname;
my $rdbh=connectdb($remote_dbname, $remote_dbuser, $remote_dbpass, $remote_dbhost, $remote_dbport);
showstats($rdbh, "Remote Database: $remote_dbname") if $stats;

exit 0 if $stats;

# check that schemas are similar
mylog("check schemas\n");

# TODO: gather and check more information with $dbh->column_info()

my %tables;
$sth=$ldbh->prepare("select * from replication.tables");
$sth->execute();
while(my $h=$sth->fetchrow_hashref())
{
    $tables{$h->{name}} = { LO_MAGIC => $h->{lo}, EMPTY_MAGIC => EMPTY_MAGIC };
    my $h2=$ldbh->selectrow_hashref("select * from $h->{name} limit 1");
    foreach my $c (keys %$h2)
    {
	delete $tables{$h->{name}}->{EMPTY_MAGIC};
	next if $c eq PRIMARY_KEY;
	$tables{$h->{name}}->{$c} = COLUMN_MAGIC;
    }
}
$sth->finish();

$sth=$rdbh->prepare("select * from replication.tables");
$sth->execute();
while(my $h=$sth->fetchrow_hashref())
{
    if(exists $tables{$h->{name}})
    {
	next if exists $tables{$h->{name}}->{EMPTY_MAGIC};
	die "table $h->{name} lo setting does not match" if $tables{$h->{name}}->{LO_MAGIC} ne $h->{lo};

	my $h2=$rdbh->selectrow_hashref("select * from $h->{name} limit 1");
	foreach my $c (keys %$h2)
	{
	    next if $c eq PRIMARY_KEY;
	    die "table $h->{name} does not match" if $tables{$h->{name}}->{$c} ne COLUMN_MAGIC;
	}
    }
    else
    {
	mylog("table $h->{name} is not included on remote host.\n");
    }
}
$sth->finish();

# get node numbers
my $q="select val from replication.config where key = 'node_num'";
my ($local_node) = $ldbh->selectrow_array($q);
die "replication not configured on local" unless defined $local_node;
my ($remote_node) = $rdbh->selectrow_array($q);
die "replication not configured on remote" unless defined $remote_node;
die "both nodes have the same node num ($remote_node)" if $local_node == $remote_node;

# get node description
$q="select val from replication.config where key = 'node_desc'";
my ($local_desc) = $ldbh->selectrow_array($q);
my ($remote_desc) = $rdbh->selectrow_array($q);

# check that nodes are configured to sync each other
mylog("check node registring\n");

sub check_node
{
    my ($dbh, $node_num, $desc) = @_;
    return unless ref $dbh;
    return unless $node_num;
    my ($c) = $dbh->selectrow_array("select count(*) from replication.nodes where num=?", undef, $node_num);
    if($c == 0)
    {
	mylog("registering Node $node_num\n");
	$dbh->do("insert into replication.nodes(num,description) values(?,?)", undef, $node_num, $desc) or die "could not register node $node_num";
	push(@after_replication, sub {
		 mylog("Set all Sequences on Node $node_num\n");
		 $dbh->do("select replication.setallseq()");
	     } );
    }
}
check_node($ldbh, $remote_node, ($remote_desc || '_nodesc_'));
check_node($rdbh, $local_node, ($local_desc || '_nodesc_'));

sub replicate
{
    my ($from, $from_num, $to, $to_num) = @_;
    return unless ref $from;
    return unless ref $to;
    return unless defined $from_num;
    return unless defined $to_num;

    mylog("replicate $from_num -> $to_num\n");

    my $fn="/tmp/$$.lo";
    END { unlink $fn if -f $fn }

    my ($r) = $to->selectrow_array("select replication.start(?)", undef, $from_num);
    die "could not start replication" unless $r == 0;

    my $sth=$from->prepare("select * from replication.log where node=? order by id");
    $sth->execute($to_num);
    my $Z = $sth->rows();
    my $z=0;
    while(my $log=$sth->fetchrow_hashref())
    {
	my $l="$from_num->$to_num ".(++$z)."/$Z ($log->{id}) ";

	my $primary_key=$log->{PRIMARY_KEY()};

	if($log->{action} eq 'INSERT')
	{
	    mylog("$l $log->{action} $log->{tablename}::$primary_key ".$log->{LO_KEY()}."\n");

	    my $h=$from->selectrow_hashref("select * from $log->{tablename} where ".PRIMARY_KEY."=?", undef, $primary_key);

	    if($h->{LO_KEY()})
	    {
		$from->func($h->{LO_KEY()}, $fn, 'lo_export') or die "could not lo_export $log->{tablename}::$primary_key ".$h->{LO_KEY()};
		mylog("  ".(-s $fn)." bytes\n");
		$h->{LO_KEY()} = $to->func($fn, 'lo_import') or die "could not lo_import $fn";
	    }

	    my $q="insert into $log->{tablename}(";
	    my @q;
	    while(my ($k,$v) = each(%$h))
	    {
		$q.="$k,";
		push(@q, $v);
	    }
	    chop $q ; # remove trailing ,
	    $q.=") values(" . ('?,' x ($#q + 1));
	    chop $q ; # remove trailing ,
	    $q.=")";
	    $to->do($q, undef, @q) or die "could not insert";
	}
	elsif($log->{action} eq 'UPDATE')
	{
	    mylog("$l $log->{action} $log->{tablename}::$primary_key ".$log->{LO_KEY()}."\n");

	    my $h=$from->selectrow_hashref("select * from $log->{tablename} where ".PRIMARY_KEY."=?", undef, $primary_key);
	    unless(defined $h->{PRIMARY_KEY()})
	    {
		mylog("  now rows found.\n");
		next;
	    }

	    my $q="update $log->{tablename} set ";
	    my @q;
	    while(my ($k,$v) = each(%$h))
	    {
		next if $k eq PRIMARY_KEY || $k eq LO_KEY;
		$q.="$k=?,";
		push(@q, $v);
	    }
	    chop $q ; # remove trailing ,
	    $q.=" where ".PRIMARY_KEY."=?";
	    push(@q, $primary_key);
#	    mylog("  $q := ".join(',', @q)."\n");
	    $to->do($q, undef, @q) or die "could not update";

	    my $from_oid=$h->{LO_KEY()};
	    if($from_oid)
	    {
		my ($from_md5) = $from->selectrow_array("select md5(?::oid)", undef, $from_oid);
		my ($to_oid, $to_md5) = $to->selectrow_array("select ".LO_KEY.", md5(".LO_KEY."::oid) from $log->{tablename} where ".PRIMARY_KEY."=?", undef, $primary_key);

		if($from_md5 ne $to_md5)
		{
		    $from->func($from_oid, $fn, 'lo_export') or die "could not lo_export $log->{tablename}::$primary_key $from_oid";
		    mylog("  ".(-s $fn)." bytes\n");
		    if($to_oid)
		    {
			# TODO: hier rsync
			mylog("  overwrite target lo\n");
			my $fd=$to->func($to_oid, $to->{pg_INV_WRITE}, 'lo_open');
			die "could not open $log->{tablename}::$primary_key $to_oid" unless defined $fd;
			open(A, $fn) or die "could not open $fn";
			my $buf;
			$to->func($fd, $buf, length($buf), 'lo_write') while sysread(A, $buf, 1024);
			close(A);
			$to->func($fd, 'lo_close');
		    }
		    else
		    {
			mylog("  create new target lo\n");
			$to_oid = $to->func($fn, 'lo_import') or die "could not lo_import $fn";
			$to->do("update $log->{tablename} set ".LO_KEY."=? where ".PRIMARY_KEY."=?", undef, $to_oid, $primary_key) or die "could not update lo oid in target";
		    }
		}
		else
		{
		    mylog("  large objects are equal\n");
		}
	    }
	}
	elsif($log->{action} eq 'DELETE')
	{
	    mylog("$l $log->{action} $log->{tablename}::$primary_key ".$log->{LO_KEY()}."\n");

	    $to->do("delete from $log->{tablename} where ".PRIMARY_KEY."=?", undef, $primary_key) or die "could not delete from $log->{tablename} #$primary_key";
	}
	elsif($log->{action} eq 'DDL')
	{
	    mylog("$l DDL: $log->{tablename}\n");
	    $to->do("select replication.ddl(?)", undef, $log->{'tablename'}) or die "could not execute DDL";
	}
	else
	{
	    warn "$l unknown action $log->{action}";
	}

	$from->do("delete from replication.log where id=?", undef, $log->{id}) or die "could not delete log #$log->{id}";
    }
    $sth->finish();

    ($r) = $to->selectrow_array("select replication.stop(?)", undef, $from_num);
    die "could not stop replication" unless $r == 0;

    $from->do("update replication.nodes set lastsync=now() where num=?", undef, $to_num) or die "could not update lastsync on from Node #$from_num";

    mylog("commit\n");
    $to->commit() or die "could not commit on destination node #$to_num";
    $from->commit() or die "could not commit on source node #$from_num";
    begin_transaction($from);
    begin_transaction($to);

    mylog("replicate $from_num -> $to_num finished\n");
}

eval {
    replicate($ldbh,$local_node,$rdbh,$remote_node);
    replicate($rdbh,$remote_node,$ldbh,$local_node);
    foreach my $c (@after_replication)
    {
	$c->() if ref($c) eq 'CODE';
    }
};
if ($@) {
    warn "Transaction aborted because $@";
    eval {
	$ldbh->rollback;
	$rdbh->rollback;
    };
    mylog("exit failure\n");
    exit 1;
}

my $exitval=0;
if($compare_db)
{
    mylog("comparing databases\n");
    while(my ($tablename, $tab) = each(%tables))
    {
	my $m='nr::text';
	while(my ($col,$typ) = each(%$tab))
	{
	    next unless $typ eq COLUMN_MAGIC;
	    my $type=get_col_type($ldbh, $tablename, $col);
	    die "column $col of $tablename have no identical type" if $type ne get_col_type($rdbh, $tablename, $col);

	    $col="array_to_string($col, ',')" if $type =~ /\[\]$/; # array
	    if($type eq 'bit')
	    {
		$col="bit_send($col)";
		$type='bytea';
	    }
	    if($type eq 'bit varying')
	    {
		$col="varbit_send($col)";
		$type='bytea';
	    }
	    if($type eq 'circle')
	    {
		$col="circle_send($col)";
		$type='bytea';
	    }
	    if($type eq 'point')
	    {
		$col="point_send($col)";
		$type='bytea';
	    }
	    if($type eq 'line')
	    {
		$col="line_send($col)";
		$type='bytea';
	    }
	    if($type eq 'lseg')
	    {
		$col="lseg_send($col)";
		$type='bytea';
	    }
	    if($type eq 'box')
	    {
		$col="box_send($col)";
		$type='bytea';
	    }
	    if($type eq 'polygon')
	    {
		$col="poly_send($col)";
		$type='bytea';
	    }
	    if($type eq 'path')
	    {
		$col="path_send($col)";
		$type='bytea';
	    }
	    $col="${col}::int" if $type eq 'boolean';
	    $col="cash_words($col)" if $type eq 'money';
	    $col="md5($col)" if $type eq 'oid' || $type eq 'bytea';
	    $m.=" || COALESCE(${col}::text, 'nil')";
	}
	# TODO: start two background processes for query execution
	my $q="select nr, md5($m) as md5 from $tablename order by nr";

	mylog(" compare $tablename with: $q\n");

	my $sthl=$ldbh->prepare($q) or die "could not prepare compare local";
	my $sthr=$rdbh->prepare($q) or die "could not prepare compare remote";
	$sthl->execute();
	$sthr->execute();
	my $l={ nr => -1, md5 => 'a' };
	my $r={ nr => -1, md5 => 'a' };
	while($l && $r)
	{
	    if($l->{nr} < $r->{nr})
	    {
		mylog(" ${tablename}::$l->{nr} is missing on remote\n");
		$l=$sthl->fetchrow_hashref();
		$exitval=2;
	    }
	    elsif($l->{nr} > $r->{nr})
	    {
		mylog(" ${tablename}::$r->{nr} is missing on local\n");
		$r=$sthr->fetchrow_hashref();
		$exitval=2;
	    }
	    else
	    {
		if($l->{md5} ne $r->{md5})
		{
		    mylog(" ${tablename}::$r->{nr} is not equal\n");
		    $exitval=2;
		}
		$l=$sthl->fetchrow_hashref();
		$r=$sthr->fetchrow_hashref();
	    }
	}
	$sthl->finish();
	$sthr->finish();
    }
}

$ldbh->disconnect();
$rdbh->disconnect();
mylog("exit success\n");
exit $exitval;

sub get_col_type
{
    my ($dbh, $tablename, $colname) = @_;
    return unless ref $dbh;
    return unless $tablename;
    return unless $colname;
    my $sth = $dbh->column_info(undef, undef, $tablename, $colname);
    my $h=$sth->fetchrow_hashref();
    $sth->finish();
    return $h->{'TYPE_NAME'};
}

sub showstats
{
    my ($dbh, $n) = @_;
    return unless ref $dbh;

    print "$n\n";
    my $a = $dbh->selectall_arrayref("select n.description, r.tablename, count(*)
from replication.log as r left outer join replication.nodes as n on (r.node = n.num)
group by n.description, r.tablename");

    foreach my $r (@$a)
    {
	print join("\t", @$r)."\n";
    }

}
