-- $Header: /code/pg_trompe/replication.sql,v 1.15 2007/09/15 16:31:36 doj Exp $

CREATE SCHEMA replication;
COMMENT ON SCHEMA replication IS 'Tabellen und Funktionen zur Replikation der Daten auf andere Nodes.';

CREATE TABLE replication.config (
  key	text not null primary key,
  val	text,
  description	text
);
COMMENT ON TABLE replication.config IS 'Enthlt einige Key/Value Paare die Funktionen steuern.';

insert into replication.config(key,val,description) values('maxnodes', '10', 'Maximale Anzahl von Nodes.');

CREATE TABLE replication.tables (
  nr	serial not null primary key,
  name	text not null unique,
  lo	boolean not null default FALSE
);
COMMENT ON TABLE replication.tables IS 'Tabellennamen die von der Replikation erfasst werden sollen.';

CREATE TABLE replication.nodes (
  num		int not null primary key,
  description	text,
  added		timestamp not null default now(),
  lastsync	timestamp
);
COMMENT ON TABLE replication.nodes IS 'Nodes mit denen synchronisiert wird.';

CREATE OR REPLACE FUNCTION replication.nodes_insert()
RETURNS trigger
AS $emp_stamp$
 DECLARE
  t RECORD;
 BEGIN
  FOR t IN select * from replication.tables order by nr LOOP

    IF t.lo = TRUE THEN
      EXECUTE 'insert into replication.log(node, tablename, action, nr, lo)
               select ''' || NEW.num || ''' , ''' || t.name || ''' , ''INSERT'' , nr , lo
               from ' || quote_ident(t.name);
    ELSE
      EXECUTE 'insert into replication.log(node, tablename, action, nr)
               select ''' || NEW.num || ''' , ''' || t.name || ''' , ''INSERT'' , nr
               from ' || quote_ident(t.name);
    END IF;

  END LOOP;
  RETURN NEW;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.nodes_insert() IS 'Wird vom Insert-Trigger auf Tabelle replication.tables aufgerufen. Schreibt INSERT msgs aller Tabellen in die replication.log Tabelle.';

CREATE TRIGGER nodes_insert_trig
AFTER INSERT
ON replication.nodes
FOR EACH ROW EXECUTE PROCEDURE replication.nodes_insert();

CREATE TABLE replication.nodes_skip (
  num		int not null primary key
);
COMMENT ON TABLE replication.nodes_skip IS 'Enthlt Node Nummern, fr die *keine* Ereignisse in replication.log eingetragen werden. Dies ist idR der eigene Node und der whrend der Replikation verwendete Node.';

CREATE TABLE replication.log (
  id		serial8 primary key,
  node		int  not null references replication.nodes on update cascade on delete cascade deferrable,
  tablename	text not null,
  action	text not null check(action = 'INSERT' or action = 'UPDATE' or action = 'DELETE' or action = 'DDL'),
  nr		int8,
  lo		int8
);
COMMENT ON TABLE replication.log IS 'Events die noch Nodes an bertragen werden mssen.';

CREATE OR REPLACE VIEW replication.nodestats AS
select node, count(*) as pending_updates, (select lastsync from replication.nodes where num=node) as lastsync
from replication.log
group by node;
COMMENT ON VIEW replication.nodestats IS 'Zeigt die Anzahl noch nicht ausgefhrter Update Events aller Nodes.';

CREATE OR REPLACE FUNCTION replication.newevent(tablename_ text, action_ text, nr_ int8, lo_ int8)
RETURNS int
AS $emp_stamp$
 BEGIN
  insert into replication.log(node,tablename,action,nr,lo)
  select num, tablename_, action_, nr_, lo_
  from replication.nodes
  where num not in (select num from replication.nodes_skip);
  RETURN 0;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.newevent(tablename_ text, action_ text, nr_ int8, lo_ int8) IS 'Schreibt einen neuen Eintrag in replication.log.';

CREATE OR REPLACE FUNCTION replication.tableevent()
RETURNS trigger
AS $emp_stamp$
 BEGIN
  IF TG_OP = 'INSERT' THEN
    perform replication.newevent(TG_RELNAME, TG_OP, NEW.nr, NULL);
    RETURN NEW;

  ELSIF TG_OP = 'UPDATE' THEN
    IF NEW.nr <> OLD.nr THEN
      RAISE EXCEPTION 'Altering PK is forbidden in replicated table % (% -> %)', TG_RELNAME, OLD.nr, NEW.nr;
      RETURN NULL;
--      perform replication.newevent(TG_RELNAME, 'DELETE', OLD.nr, NULL);
--      perform replication.newevent(TG_RELNAME, 'INSERT', NEW.nr, NULL);
    ELSE
      delete from replication.log where tablename=TG_RELNAME and action=TG_OP and nr=NEW.nr;
      perform replication.newevent(TG_RELNAME, TG_OP, NEW.nr, NULL);
    END IF;
    RETURN NEW;

  ELSIF TG_OP = 'DELETE' THEN
    delete from replication.log where tablename=TG_RELNAME and nr=OLD.nr;
    perform replication.newevent(TG_RELNAME, TG_OP, OLD.nr, NULL);
    RETURN OLD;
  END IF;

  RETURN NULL;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.tableevent() IS 'Trigger-Funktion fr Tabellen ohne Large Objects. Schreibt mit newevent() in replication.log.';

CREATE OR REPLACE FUNCTION replication.loevent()
RETURNS trigger
AS $emp_stamp$
 BEGIN
  IF TG_OP = 'INSERT' THEN
    perform replication.newevent(TG_RELNAME, TG_OP, NEW.nr, NEW.lo::int8);
    RETURN NEW;

  ELSIF TG_OP = 'UPDATE' THEN
    IF NEW.nr <> OLD.nr THEN
      RAISE EXCEPTION 'Altering PK is forbidden in replicated table % (% -> %)', TG_RELNAME, OLD.nr, NEW.nr;
      RETURN NULL;
    END IF;
    IF OLD.lo IS NULL and NEW.lo IS NULL THEN
      delete from replication.log where tablename=TG_RELNAME and action=TG_OP and nr=NEW.nr;
      perform replication.newevent(TG_RELNAME, TG_OP, NEW.nr, NULL);
    ELSEIF OLD.lo IS NULL THEN
      perform replication.newevent(TG_RELNAME, 'INSERT', NEW.nr, NEW.lo::int8);
    ELSEIF NEW.lo IS NULL THEN
      perform lo_unlink(OLD.lo);
      perform replication.newevent(TG_RELNAME, 'DELETE', OLD.nr, OLD.lo::int8);
    ELSIF NEW.lo <> OLD.lo THEN
      perform lo_unlink(OLD.lo);
      perform replication.newevent(TG_RELNAME, 'DELETE', OLD.nr, OLD.lo::int8);
      perform replication.newevent(TG_RELNAME, 'INSERT', NEW.nr, NEW.lo::int8);
    ELSE
      delete from replication.log where tablename=TG_RELNAME and action=TG_OP and nr=NEW.nr;
      perform replication.newevent(TG_RELNAME, TG_OP, NEW.nr, NEW.lo::int8);
    END IF;

    RETURN NEW;

  ELSIF TG_OP = 'DELETE' THEN
    delete from replication.log where tablename=TG_RELNAME and nr=OLD.nr;
    perform lo_unlink(OLD.lo);
    perform replication.newevent(TG_RELNAME, TG_OP, OLD.nr, OLD.lo::int8);

    RETURN OLD;

  END IF;

  RETURN NULL;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.loevent() IS 'Trigger-Funktion fr Tabellen mit Large Objects. Schreibt mit newevent() in replication.log.';

CREATE OR REPLACE FUNCTION replication.setseq(tablename text)
RETURNS int
AS $emp_stamp$
 DECLARE
  node_num text;
  maxnodes text;
  seq text;
 BEGIN
  -- check if replication is set up correctly
  SELECT val INTO node_num from replication.config where key = 'node_num';
  IF NOT FOUND THEN
    RAISE EXCEPTION 'replication node number not set';
    RETURN -1;
  END IF;

  -- get maxnodes
  SELECT val INTO maxnodes from replication.config where key = 'maxnodes';
  IF NOT FOUND THEN
    RAISE EXCEPTION 'replication maxnodes not set';
    RETURN -1;
  END IF;

  -- get new PK value. Because we can not SELECT INTO in conjuction
  -- with EXECUTE we use the replication.config table here with our
  -- temporary key 'DOJ'
  EXECUTE 'insert into replication.config(key,val)
  select ''DOJ'', COALESCE(max(nr)+' || maxnodes || ',' || node_num || ')
  from ' || tablename || '
  where (nr % ' || maxnodes || ') = ' || node_num;

  SELECT val INTO seq from replication.config where key = 'DOJ';
  IF NOT FOUND THEN
    RAISE EXCEPTION 'Error when inserting special DOJ config';
    RETURN -1;
  END IF;
  delete from replication.config where key = 'DOJ';

  -- set sequence value
  EXECUTE 'ALTER SEQUENCE ' || tablename || '_nr_seq
  INCREMENT BY ' || maxnodes || '
  NO MINVALUE NO MAXVALUE
  RESTART WITH ' || seq || '
  NO CYCLE;';

  RETURN 1;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.setseq(tablename text) IS 'Setzt die Primary Key Sequence fr Tabelle tablename neu.';

CREATE OR REPLACE FUNCTION replication.setallseq()
RETURNS int
AS $emp_stamp$
 DECLARE
  t RECORD;
 BEGIN
  FOR t IN select * from replication.tables LOOP
    PERFORM replication.setseq(t.name);
  END LOOP;
  RETURN 0;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.setallseq() IS 'Setzt alle PK Sequences der registrierten Tabellen neu.';

CREATE OR REPLACE FUNCTION replication._register(tablename text)
RETURNS int
AS $emp_stamp$
 DECLARE
  c int;
 BEGIN
  -- check if table is not already handled
  SELECT count(*) INTO c from replication.tables where name = tablename;
  IF c > 0 THEN
    RAISE EXCEPTION 'table % already in replication', tablename;
    RETURN -1;
  END IF;

  -- register table
  insert into replication.tables(name) values(tablename);

  SELECT INTO c replication.setseq(tablename);

  RETURN c;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication._register(tablename text) IS 'Bereitet die Replikationskonfiguration fr eine neue Tabelle vor.';

CREATE OR REPLACE FUNCTION replication.register(tablename text)
RETURNS text
AS $emp_stamp$
 DECLARE
  i int;
 BEGIN
  -- register the table
  SELECT INTO i replication._register(tablename);
  IF i <> 1 THEN
    RETURN NULL;
  END IF;

  -- create event trigger on table
  EXECUTE 'CREATE TRIGGER ' || tablename || '_replication_event
           BEFORE INSERT or UPDATE or DELETE
           ON ' || tablename || '
           FOR EACH ROW EXECUTE PROCEDURE replication.tableevent()';

  RETURN tablename;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.register(tablename text) IS 'Registriert eine Tabelle ohne Large-Objects bei der Replikationskonfiguration.';

CREATE OR REPLACE FUNCTION replication.registerlo(tablename text)
RETURNS text
AS $emp_stamp$
 DECLARE
  i int;
 BEGIN
  -- register the table
  SELECT INTO i replication._register(tablename);
  IF i <> 1 THEN
    RETURN NULL;
  END IF;

  update replication.tables set lo=TRUE where name=tablename;

  -- create event trigger to large object on table
  EXECUTE 'CREATE TRIGGER ' || tablename || '_replication_loevent
           BEFORE INSERT or UPDATE or DELETE
           ON ' || tablename || '
           FOR EACH ROW EXECUTE PROCEDURE replication.loevent()';

  RETURN tablename;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.registerlo(tablename text) IS 'Registriert eine Tabelle mit Large-Objects bei der Replikationskonfiguration.';

CREATE OR REPLACE FUNCTION replication.node_num(num_ int, d text)
RETURNS int
AS $emp_stamp$
 DECLARE
  n int;
  maxnodes int;
 BEGIN
  -- get maxnodes
  SELECT val::int INTO maxnodes from replication.config where key = 'maxnodes';

  -- check node number
  IF num_ < 0 or num_ >= maxnodes THEN
    RAISE NOTICE 'Node Num % is outside of configured range [0..%)', num_, maxnodes;
    RETURN -1;
  END IF;

  -- check if already configured
  SELECT val::int INTO n from replication.config where key='node_num';
  IF FOUND THEN
    RAISE NOTICE 'This Node is already configured as #%', n;
    RETURN -1;
  END IF;

  insert into replication.config(key,val,description) values('node_num', num_, 'Die Node Nummber dieses Nodes.');
  insert into replication.config(key,val,description) values('node_desc', d, 'Beschreibung dieses Nodes.');
  insert into replication.nodes_skip(num) values(num_);
--  delete from replication.log where node=num_;

  RETURN num_;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.node_num(num_ int, d text) IS 'Setzt die Node Nummer.';

CREATE OR REPLACE FUNCTION replication.drop_node(n int)
RETURNS int
AS $emp_stamp$
 BEGIN
  delete from replication.nodes where num=n;
  IF FOUND THEN
    RETURN n;
  END IF;
  RAISE NOTICE 'Node % was not included in replication.', n;
  RETURN -1;
 END;
$emp_stamp$ LANGUAGE plpgsql;
COMMENT ON FUNCTION replication.drop_node(n int) IS 'Entfernt einen Node aus der Replikation.';

CREATE OR REPLACE FUNCTION md5(id oid)
RETURNS text
AS $$
 DECLARE
  fd        integer;
  size      integer;
  hashval   text;
  INV_READ  constant integer := x'40000'::integer; -- from libpq-fs.h
  SEEK_SET  constant integer := 0;
  SEEK_END  constant integer := 2;
 BEGIN
  fd   := lo_open(id, INV_READ);
  size := lo_lseek(fd, 0, SEEK_END);
  PERFORM lo_lseek(fd, 0, SEEK_SET);
  hashval := md5(loread(fd, size));
  PERFORM lo_close(fd);
  RETURN hashval;
 END;
$$
LANGUAGE plpgsql stable strict;
COMMENT ON FUNCTION md5(id oid) IS 'Calculates the md5 sum of a large object.';

CREATE OR REPLACE FUNCTION replication.ddl(t text)
RETURNS int
AS $$
 DECLARE
  r int;
 BEGIN
  SELECT INTO r replication.newevent(t, 'DDL', NULL, NULL);
  IF r = 0 THEN
    EXECUTE t;
  END IF;
  RETURN r;
 END;
$$
LANGUAGE plpgsql strict;
COMMENT ON FUNCTION replication.ddl(t text) IS 'Store a DDL commands in the log table.';

CREATE OR REPLACE FUNCTION replication.start(remote_num int)
RETURNS int
AS $$
 DECLARE
  n int;
 BEGIN
  SELECT INTO n val::int from replication.config where key='node_num';
  IF NOT FOUND THEN
    RAISE EXCEPTION 'Replication has not been set up on this node.';
    RETURN -1;
  END IF;
  IF n = remote_num THEN
    RAISE EXCEPTION 'You can''t replicate onto yourself';
    RETURN -1;
  END IF;

  insert into replication.nodes_skip(num) values(remote_num);

  RETURN 0;
 END;
$$
LANGUAGE plpgsql strict;
COMMENT ON FUNCTION replication.start(remote_num int) IS 'Startet die Replikation mit Node remote_num.';

CREATE OR REPLACE FUNCTION replication.stop(remote_num int)
RETURNS int
AS $$
 DECLARE
  n int;
 BEGIN
  SELECT INTO n val::int from replication.config where key='node_num';
  IF NOT FOUND THEN
    RAISE EXCEPTION 'Replication has not been set up on this node.';
    RETURN -1;
  END IF;
  IF n = remote_num THEN
    RAISE EXCEPTION 'You can''t replicate onto yourself';
    RETURN -1;
  END IF;

  delete from replication.nodes_skip where num=remote_num;

  RETURN 0;
 END;
$$
LANGUAGE plpgsql strict;
COMMENT ON FUNCTION replication.stop(remote_num int) IS 'Beendet die Replikation mit Node remote_num.';
